import os
import pathlib
from typing import Union

import numpy

PathLike = Union[str, bytes, pathlib.Path, os.PathLike]
Number = Union[int, float]
Value = Union[Number, bool]

NumpyNumber = Union[numpy.floating, numpy.integer]
NumpyValue = Union[NumpyNumber, numpy.bool_]

NumpySupportedNumber = Union[Number, NumpyNumber]
NumpySupportedFloat = Union[float, numpy.floating]
NumpySupportedValue = Union[Value, NumpyValue]



from typing import (
    Any,
    Set,
    Type,
    Union,
    TypeVar,
    Iterable,
    Optional,
    Sequence,
    Collection,
    SupportsInt,
    SupportsFloat,
    SupportsIndex,
    cast,
)

import numpy as np
import numpy.typing

from qm.type_hinting import Value
from qm.exceptions import QmValueError

GeneralConversionType = Union[str, bytes, bytearray, memoryview]
FloatConversionType = Union[SupportsFloat, SupportsIndex, GeneralConversionType]
IntConversionType = Union[SupportsInt, SupportsIndex, GeneralConversionType]
Bool = Union[bool, np.bool_]

T = TypeVar("T", IntConversionType, FloatConversionType, Bool)


def convert_object_type(obj: T) -> Value:
    if isinstance(obj, (np.bool_, bool)):
        return bool(obj)
    if isinstance(obj, (np.integer, int)):
        return int(obj)
    if isinstance(obj, (np.floating, float)):
        return float(obj)
    raise QmValueError(f"cannot convert {type(obj)} to int | float | bool")


def get_all_iterable_data_types(it: Iterable[Any]) -> Set[Type[Any]]:
    return {type(e) for e in it}


C = TypeVar("C")


def collection_has_type(collection: Collection[C], type_to_check: Type[C], include_subclasses: bool) -> bool:
    if include_subclasses:
        return any([isinstance(i, type_to_check) for i in collection])
    else:
        return any([type(i) is type_to_check for i in collection])


def collection_has_type_bool(collection: Collection[C]) -> bool:
    return collection_has_type(collection, bool, False) or collection_has_type(collection, np.bool_, True)


def collection_has_type_int(collection: Collection[C]) -> bool:
    return collection_has_type(collection, int, False) or collection_has_type(collection, np.integer, True)


def collection_has_type_float(collection: Collection[C]) -> bool:
    return collection_has_type(collection, float, False) or collection_has_type(collection, np.floating, True)


def is_iter(x: Any) -> bool:
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


def get_iterable_elements_datatype(it: Union[numpy.typing.NDArray[Any], Sequence[Any], Any]) -> Optional[Type[Any]]:
    if isinstance(it, np.ndarray):
        return type(it[0].item())
    elif is_iter(it):
        sequence = cast(Sequence[Any], it)
        if len(get_all_iterable_data_types(sequence)) > 1:
            raise ValueError("Multiple datatypes encountered in iterable object")
        if isinstance(sequence[0], np.generic):
            return type(sequence[0].item())
        else:
            return type(sequence[0])
    else:
        return None



import json
import warnings
from typing import List, Tuple, Union, Literal, Optional, Sequence, overload

import betterproto

from qm.api.v2.job_api import JobApi
from qm.octave import QmOctaveConfig
from qm.persistence import BaseStore
from qm.program.program import Program
from qm.utils import deprecation_message
from qm.exceptions import FunctionInputError
from qm.jobs.job_queue_mock import QmQueueMock
from qm.utils.config_utils import get_fem_config
from qm.octave.octave_manager import OctaveManager
from qm.simulate.interface import SimulationConfig
from qm.api.v2.job_api.job_api import OldJobApiMock
from qm.type_hinting.config_types import DictQuaConfig
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.api.v2.qm_api import QmApi, IoValue, NoRunningQmJob
from qm.type_hinting import Value, Number, NumpySupportedValue
from qm.type_hinting.general import PathLike, NumpySupportedFloat
from qm.api.v2.job_api.simulated_job_api import OldSimulatedJobApiMock
from qm.api.models.compiler import CompilerOptionArguments, standardize_compiler_params
from qm.grpc.qua_config import QuaConfig, QuaConfigMicrowaveFemDec, QuaConfigAdcPortReference, QuaConfigDacPortReference


class OldQmApiMock(QmApi):
    SIMULATED_JOB_CLASS = OldSimulatedJobApiMock

    def __init__(
        self,
        connection_details: ConnectionDetails,
        qm_id: str,
        store: BaseStore,
        capabilities: ServerCapabilities,
        octave_config: Optional[QmOctaveConfig],
        octave_manager: OctaveManager,
        pb_config: Optional[QuaConfig] = None,
    ):
        super().__init__(connection_details, qm_id, store, capabilities, octave_config, octave_manager, pb_config)
        self._queue = QmQueueMock(store=store, api=self, capabilities=self._caps)

    def _get_job(self, job_id: str, store: BaseStore) -> OldJobApiMock:
        return OldJobApiMock(self.connection_details, job_id, store, capabilities=self._caps)

    def get_job_by_id(self, job_id: str) -> OldJobApiMock:
        warnings.warn(
            deprecation_message(
                method="qm.get_job_by_id",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please use `qm.get_job()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._get_job(job_id, self._store)

    @property
    def queue(self) -> QmQueueMock:
        warnings.warn(
            deprecation_message(
                method="qm.queue",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This property is going to be removed, all functionality will exist directly under "
                "`QuantumMachine`. For example, instead of `qm.queue.add(prog)` use `qm.add_to_queue(prog)`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._queue

    def close(self) -> bool:  # type: ignore[override]
        super().close()
        return True

    def execute(
        self,
        program: Program,
        duration_limit: None = None,
        data_limit: None = None,
        force_execution: None = None,
        dry_run: None = None,
        simulate: Optional[SimulationConfig] = None,
        *,
        compiler_options: Optional[CompilerOptionArguments] = None,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> JobApi:
        if not isinstance(program, Program):
            raise Exception("program argument must be of type qm.program.Program")

        for x, name in [
            (duration_limit, "`duration_limit'"),
            (data_limit, "`data_limit'"),
            (force_execution, "`force_execution'"),
            (dry_run, "`dry_run'"),
        ]:
            if x is not None:
                warnings.warn(
                    deprecation_message(
                        method=f"The argument {name}",
                        deprecated_in="1.1.8",
                        removed_in="1.2.0",
                    ),
                )

        compiler_options = standardize_compiler_params(compiler_options, strict, flags)

        if simulate is not None:
            warnings.warn(
                deprecation_message(
                    method="The argument simulate",
                    deprecated_in="1.1.8",
                    removed_in="1.2.0",
                    details="The simulate argument is deprecated, please use the simulate method.",
                ),
            )
            return self.simulate(program, simulate, compiler_options=compiler_options)

        self.clear_queue()
        current_running_job = self._get_running_job()
        if current_running_job is not None:
            current_running_job.cancel()

        new_job_api = super().execute(program, compiler_options=compiler_options)
        new_job_api.wait_until({"Running", "Processing"}, timeout=5 * 60)
        # The timeout here is just for the backwards compatibility behaviour.
        # See that in the father function there is no `wait_until`
        return new_job_api

    def list_controllers(self) -> Tuple[str, ...]:
        warnings.warn(
            deprecation_message(
                method="qm.list_controllers",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please get data from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        return tuple(config.v1_beta.control_devices) or tuple(config.v1_beta.controllers)

    def set_mixer_correction(
        self,
        mixer: str,
        intermediate_frequency: Number,
        lo_frequency: Number,
        values: Tuple[float, float, float, float],
    ) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_mixer_correction",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please use `job.set_element_correction()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config: DictQuaConfig = {
            "version": 1,
            "mixers": {
                mixer: [
                    {
                        "correction": values,
                        "intermediate_frequency": intermediate_frequency,
                        "lo_frequency": lo_frequency,
                    }
                ]
            },
        }
        self.update_config(config)
        job = self._get_running_job()
        if job is not None:
            pb_config = self._get_pb_config()
            for name, element_config in pb_config.v1_beta.elements.items():
                if betterproto.serialized_on_wire(element_config.mix_inputs):
                    if element_config.mix_inputs.mixer == mixer:
                        if element_config.intermediate_frequency == intermediate_frequency:
                            if element_config.mix_inputs.lo_frequency == lo_frequency:
                                job.set_element_correction(name, values)

    def set_intermediate_frequency(self, element: str, freq: float) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_intermediate_frequency",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use "
                "`job.set_intermediate_frequency()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        job.set_intermediate_frequency(element, freq)

    def get_intermediate_frequency(self, element: str) -> float:
        warnings.warn(
            deprecation_message(
                method="qm.get_intermediate_frequency",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use "
                "`job.get_intermediate_frequency()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_intermediate_frequency(element)

    def get_output_dc_offset_by_element(
        self, element: str, iq_input: Optional[Literal["I", "Q", "single"]] = None
    ) -> float:
        warnings.warn(
            deprecation_message(
                method="qm.get_output_dc_offset_by_element",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please get idle value from `qm.get_config()`"
                " or current value from job `job.get_output_dc_offset_by_element()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_output_dc_offset_by_element(element, iq_input)

    @overload
    def set_output_dc_offset_by_element(self, element: str, input: Literal["single", "I", "Q"], offset: float) -> None:
        pass

    @overload
    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Tuple[Literal["I", "Q"], Literal["I", "Q"]],
        offset: Union[Tuple[float, float], List[float]],
    ) -> None:
        pass

    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Union[Literal["single", "I", "Q"], Tuple[Literal["I", "Q"], Literal["I", "Q"]], List[Literal["I", "Q"]]],
        offset: Union[float, Tuple[float, float], List[float]],
    ) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_output_dc_offset_by_element",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please set idle value with `qm.update_config()`"
                " or current value from job `job.set_output_dc_offset_by_element()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        if isinstance(input, str):
            if not isinstance(offset, (int, float)):
                raise FunctionInputError(f"Input should be int or float, got {type(offset)}")
            if input in {"I", "Q"}:
                ports = self._get_input_ports_from_mixed_input_element(element)
                port = ports[0] if input == "I" else ports[1]
            else:
                port = self._get_input_port_from_single_input_element(element)
            config = self._create_config_for_output_dc_offset_setting(port, offset)
            self.update_config(config)
            job = self._get_running_job()
            if job is not None:
                job.set_output_dc_offset_by_element(element, input, offset)
        elif isinstance(input, (list, tuple)):
            if not set(input) <= {"I", "Q"}:
                raise FunctionInputError(f"Input names should be 'I' or 'Q', got {input}")
            if not (isinstance(offset, (list, tuple)) and len(input) == len(offset)):
                raise FunctionInputError(
                    f"input should be two iterables of the same size," f"got input = {input} and offset = {offset}"
                )
            ports = self._get_input_ports_from_mixed_input_element(element)
            for _input, _offset in zip(input, offset):
                _port = ports[0] if _input == "I" else ports[1]
                config = self._create_config_for_output_dc_offset_setting(_port, _offset)
                self.update_config(config)
                job = self._get_running_job()
                if job is not None:
                    job.set_output_dc_offset_by_element(element, input, offset)
        else:
            raise FunctionInputError(f"Input should be str or tuple, got {type(input)}")

    def set_output_filter_by_element(
        self,
        element: str,
        input: str,
        feedforward: Sequence[NumpySupportedFloat],
        feedback: Sequence[NumpySupportedFloat],
    ) -> None:
        raise NotImplementedError

    def set_input_dc_offset_by_element(self, element: str, output: str, offset: float) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_input_dc_offset_by_element",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use "
                "`job.set_input_dc_offset_by_element()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        port = self._get_output_port_from_element(element, output)
        config = self._create_config_for_input_dc_offset_setting(port, offset)
        self.update_config(config)

        job = self._get_running_job()
        if job is not None:
            job.set_input_dc_offset_by_element(element, output, offset)

    def get_input_dc_offset_by_element(self, element: str, output: str) -> float:
        warnings.warn(
            deprecation_message(
                method="qm.get_input_dc_offset_by_element",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please get the value from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        port = self._get_output_port_from_element(element, output)
        fem_config = get_fem_config(config, port)
        if isinstance(fem_config, QuaConfigMicrowaveFemDec):
            raise ValueError(f"Element {element} does not support dc offset.")
        return fem_config.analog_inputs[port.number].offset

    def get_digital_delay(self, element: str, digital_input: str) -> int:
        warnings.warn(
            deprecation_message(
                method="qm.get_digital_delay",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use use "
                "`job.get_output_digital_delay()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_output_digital_buffer(element, digital_input)

    def set_digital_delay(self, element: str, digital_input: str, delay: int) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_digital_delay",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use use "
                "`job.set_output_digital_delay()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        job.set_output_digital_delay(element, digital_input, delay)

    def get_digital_buffer(self, element: str, digital_input: str) -> int:
        warnings.warn(
            deprecation_message(
                method="qm.get_digital_buffer",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use use "
                "`job.get_output_digital_buffer()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_output_digital_buffer(element, digital_input)

    def set_digital_buffer(self, element: str, digital_input: str, buffer: int) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_digital_buffer",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use use "
                "`job.set_output_digital_buffer()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        job.set_output_digital_buffer(element, digital_input, buffer)

    def get_time_of_flight(self, element: str) -> int:
        warnings.warn(
            deprecation_message(
                method="qm.get_time_of_flight",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please get the value from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        tof = config.v1_beta.elements[element].time_of_flight
        if tof is None:
            raise ValueError(f"Time of flight for element {element} is not set")
        return tof

    def get_smearing(self, element: str) -> int:
        warnings.warn(
            deprecation_message(
                method="qm.get_smearing",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please get the value from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        smearing = config.v1_beta.elements[element].smearing
        if smearing is None:
            raise ValueError(f"Smearing for element {element} is not set")
        return smearing

    @property
    def io1(self) -> IoValue:
        warnings.warn(
            deprecation_message(
                method="qm.io1",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This property is going to be removed, please use `job.get_io_values()[0]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io1_value()

    @io1.setter
    def io1(self, value: Value) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.io1",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This property is going to be removed, please use `job.set_io_values(io1=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io1_value(value)

    @property
    def io2(self) -> IoValue:
        warnings.warn(
            deprecation_message(
                method="qm.io2",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This property is going to be removed, please use `job.get_io_values()[1]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io1_value()

    @io2.setter
    def io2(self, value: Value) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.io2",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This property is going to be removed, please use `job.set_io_values(io2=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io2_value(value)

    def set_io1_value(self, value_1: Value) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_io1_value",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use `job.set_io_values(io1=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io_values(value_1=value_1)

    def set_io2_value(self, value_2: Value) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_io2_value",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use `job.set_io_values(io2=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io_values(value_2=value_2)

    def set_io_values(
        self,
        value_1: Optional[NumpySupportedValue] = None,
        value_2: Optional[NumpySupportedValue] = None,
    ) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.set_io_values",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use `job.set_io_values()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        if value_1 is None and value_2 is None:
            return

        job = self._strict_get_running_job()
        job.set_io_values(value_1, value_2)

    def get_io1_value(self) -> IoValue:
        warnings.warn(
            deprecation_message(
                method="qm.get_io2_value",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use `job.get_io_values()[0]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io_values()[0]

    def get_io2_value(self) -> IoValue:
        warnings.warn(
            deprecation_message(
                method="qm.get_io2_value",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use `job.get_io_values()[1]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io_values()[1]

    def get_io_values(self) -> List[IoValue]:
        warnings.warn(
            deprecation_message(
                method="qm.get_io_values",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be moved to the job API, please use `job.get_io_values()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        running_job = self._strict_get_running_job()
        resp1, resp2 = running_job.get_io_values()
        return [
            {
                "io_number": 1,
                "int_value": resp1.int_value,
                "fixed_value": resp1.double_value,
                "boolean_value": resp1.boolean_value,
            },
            {
                "io_number": 2,
                "int_value": resp2.int_value,
                "fixed_value": resp2.double_value,
                "boolean_value": resp2.boolean_value,
            },
        ]

    def save_config_to_file(self, filename: PathLike) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.save_config_to_file",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        with open(filename, "w") as writer:
            json.dump(self.get_config(), writer)

    def get_running_job(self) -> Optional[OldJobApiMock]:
        warnings.warn(
            deprecation_message(
                method="qm.get_running_job",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method is going to be removed, please use `qm.get_jobs(status=['Running'])`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._get_running_job()

    def _get_running_job(self) -> Optional[OldJobApiMock]:
        jobs = self.get_jobs(status=["Running"])
        if jobs:
            return self._get_job(jobs[0].id, store=self._store)
        return None

    def _strict_get_running_job(self) -> OldJobApiMock:
        job = self._get_running_job()
        if job is None:
            raise NoRunningQmJob("No running job found")
        return job

    def _get_output_port_from_element(self, element_name: str, port_name: str) -> QuaConfigAdcPortReference:
        element = self._get_pb_config().v1_beta.elements[element_name]
        return element.multiple_outputs.port_references[port_name]

    def _get_input_ports_from_mixed_input_element(
        self, element_name: str
    ) -> Tuple[QuaConfigDacPortReference, QuaConfigDacPortReference]:
        element = self._get_pb_config().v1_beta.elements[element_name]
        return element.mix_inputs.i, element.mix_inputs.q

    def _get_input_port_from_single_input_element(self, element_name: str) -> QuaConfigDacPortReference:
        element = self._get_pb_config().v1_beta.elements[element_name]
        return element.single_input.port

    @staticmethod
    def _create_config_for_input_dc_offset_setting(port: QuaConfigAdcPortReference, value: Number) -> DictQuaConfig:
        return {
            "version": 1,
            "controllers": {port.controller: {"fems": {port.fem: {"analog_inputs": {port.number: {"offset": value}}}}}},
        }

    @staticmethod
    def _create_config_for_output_dc_offset_setting(port: QuaConfigDacPortReference, value: Number) -> DictQuaConfig:
        return {
            "version": 1,
            "controllers": {
                port.controller: {"fems": {port.fem: {"analog_outputs": {port.number: {"offset": value}}}}}
            },
        }



from qm.program.program import Program
from qm.utils.deprecation_utils import throw_warning

_Program = Program

throw_warning(
    "'qm.program._Program' is moved as of 1.1.0 and will be removed in 1.2.0. " "use 'qm.Program' instead",
    category=DeprecationWarning,
    stacklevel=2,
)



from qm.utils.deprecation_utils import throw_warning
from qm.quantum_machines_manager import QuantumMachinesManager  # noqa

throw_warning(
    "'qm.QuantumMachinesManager.QuantumMachinesManager' is moved as of 1.1.2 and will be removed in 1.2.0. "
    "use 'qm.QuantumMachinesManager' instead",
    category=DeprecationWarning,
    stacklevel=2,
)



import datetime
import warnings
from typing import List, Optional, Protocol

import betterproto

from qm.type_hinting import Value
from qm.persistence import BaseStore
from qm.exceptions import QmQuaException
from qm.api.frontend_api import FrontendApi
from qm.grpc.frontend import JobExecutionStatus
from qm.api.models.capabilities import ServerCapabilities
from qm.api.job_manager_api import create_job_manager_from_api
from qm.utils import deprecation_message, deprecate_to_property


class JobStateProtocol(Protocol):
    added_by: Optional[str]
    time_added: Optional[datetime.datetime]


class QmBaseJob:
    def __init__(
        self,
        job_id: str,
        machine_id: str,
        frontend_api: FrontendApi,
        capabilities: ServerCapabilities,
        store: BaseStore,
    ):
        self._id = job_id
        self._machine_id = machine_id
        self._frontend = frontend_api
        self._capabilities = capabilities
        self._store = store

        self._job_manager = create_job_manager_from_api(frontend_api)

        self._added_user_id: Optional[str] = None
        self._time_added: Optional[datetime.datetime] = None
        self._initialize_from_job_status()

    def _initialize_from_job_status(self) -> None:
        status: JobExecutionStatus = self._job_manager.get_job_execution_status(self._id, self._machine_id)
        name, one_of = betterproto.which_one_of(status, "status")

        if name in ("pending", "running", "completed", "loading") and one_of is not None:
            job_state: JobStateProtocol = one_of
            self._added_user_id = job_state.added_by
            self._time_added = job_state.time_added

    @property
    def status(self) -> str:
        """Returns the status of the job, one of the following strings:
        "unknown", "pending", "running", "completed", "canceled", "loading", "error"
        """
        status: JobExecutionStatus = self._job_manager.get_job_execution_status(self._id, self._machine_id)
        name, _ = betterproto.which_one_of(status, "status")
        return name

    @property
    def id(self) -> str:
        """Returns:
        The id of the job
        """
        class_name = self.__class__.__name__
        return deprecate_to_property(
            self._id,
            f"{class_name}.id()",
            "1.1.0",
            "1.2.0",
            f"id moved to be property, use '{class_name}.id' instead",
        )

    @property
    def user_added(self) -> Optional[str]:
        return self._added_user_id

    @property
    def time_added(self) -> Optional[datetime.datetime]:
        return self._time_added

    def insert_input_stream(
        self,
        name: str,
        data: List[Value],
    ) -> None:
        warnings.warn(
            deprecation_message(
                method="job.insert_input_stream",
                deprecated_in="1.1.8",
                removed_in="1.2.0",
                details="This method was renamed to `job.push_to_input_stream`.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        self.push_to_input_stream(name, data)

    def push_to_input_stream(self, name: str, data: List[Value]) -> None:
        """Insert data to the input stream declared in the QUA program.
        The data is then ready to be read by the program using the advance
        input stream QUA statement.

        Multiple data entries can be inserted before the data is read by the program.

        See [Input streams](/qm-qua-sdk/docs/Guides/features/#input-streams) for more information.

        -- Available from QOP 2.0 --

        Args:
            name: The input stream name the data is to be inserted to.
            data: The data to be inserted. The data's size must match
                the size of the input stream.
        """
        if not self._capabilities.supports_input_stream:
            raise QmQuaException("`insert_input_stream()` is not supported by the QOP version.")

        if not isinstance(data, list):
            data = [data]

        self._job_manager.insert_input_stream(self.id, name, data)



import time
import logging
from pathlib import Path
from dataclasses import asdict, dataclass
from typing import Dict, Tuple, Union, Optional, cast

import numpy as np
from tinydb.table import Document
from tinydb import Query, TinyDB, where
from tinydb.storages import JSONStorage

from qm.type_hinting.general import Number
from qm.octave.octave_mixer_calibration import MixerCalibrationResults

logger = logging.getLogger(__name__)
Correction = Tuple[float, float, float, float]


def convert_to_correction(gain: float, phase: float) -> Correction:
    s = phase
    c = np.polyval([-3.125, 1.5, 1], s**2)
    g_plus = np.polyval([0.5, 1, 1], gain)
    g_minus = np.polyval([0.5, -1, 1], gain)

    c00 = float(g_plus * c)
    c01 = float(g_plus * s)
    c10 = float(g_minus * s)
    c11 = float(g_minus * c)

    return c00, c01, c10, c11


@dataclass
class _Mode:
    octave_name: str
    octave_channel: int


@dataclass
class _LOMode:
    mode_id: int
    lo_freq: float
    gain: Optional[float]
    latest: int


@dataclass
class _IFMode:
    lo_mode_id: int
    if_freq: float
    latest: int


@dataclass
class LOCalibrationDBSchema:
    i0: float
    q0: float
    lo_mode_id: int
    dc_gain: float
    dc_phase: float
    temperature: float
    timestamp: float
    method: str

    @property
    def dc_correction(self) -> Correction:
        return convert_to_correction(self.dc_gain, self.dc_phase)


@dataclass
class IFCalibrationDBSchema:
    if_mode_id: int
    gain: float
    phase: float
    temperature: float
    timestamp: float
    method: str

    @property
    def correction(self) -> Correction:
        return convert_to_correction(self.gain, self.phase)


class ModeNotFoundError(KeyError):
    def __init__(self, query: Tuple[Union[str, Number], ...]):
        self._query = query

    def __str__(self) -> str:
        return f"Didn't find mode for {self._query}"


class CalibrationDB:
    def __init__(self, path: Union[Path, str]) -> None:
        self._file_path: Path = Path(path) / "calibration_db.json"
        self._db = TinyDB(self._file_path, indent=4, separators=(",", ": "), storage=JSONStorage)

    def __del__(self) -> None:
        self._db.close()

    def reset(self) -> None:
        self._db.close()

        self._file_path.unlink()
        self._db = TinyDB(self._file_path, indent=4, separators=(",", ": "), storage=JSONStorage)

    @property
    def file_path(self) -> str:
        return str(self._file_path)

    def _query_mode(self, octave_channel: Tuple[str, int]) -> Optional[Document]:
        return cast(
            Optional[Document],
            self._db.table("modes").get(
                (Query().octave_name == octave_channel[0]) & (Query().octave_channel == octave_channel[1])
            ),
        )

    def _get_timestamp(self, doc: Document) -> float:
        a = cast(Document, self._db.table("lo_cal").get(doc_id=doc["latest"]))
        return cast(float, a["timestamp"])

    def _query_lo_mode(self, mode_id: int, lo_freq: Number, gain: Optional[float]) -> Optional[Document]:
        if gain is not None:
            return cast(
                Optional[Document],
                self._db.table("lo_modes").get(
                    (Query().mode_id == mode_id) & (Query().lo_freq == lo_freq) & (Query().gain == gain)
                ),
            )

        lo_modes = self._db.table("lo_modes").search((Query().mode_id == mode_id) & (Query().lo_freq == lo_freq))
        if not lo_modes:
            return None

        return max(lo_modes, key=self._get_timestamp)

    def query_if_mode(self, lo_mode_id: int, if_freq: float) -> Optional[Document]:
        return cast(
            Optional[Document],
            self._db.table("if_modes").get((Query().lo_mode_id == lo_mode_id) & (Query().if_freq == if_freq)),
        )

    def _mode_id(self, octave_channel: Tuple[str, int], create: bool = False) -> int:
        query_result = self._query_mode(octave_channel)
        if query_result is None:
            if create:
                return self._db.table("modes").insert(asdict(_Mode(*octave_channel)))
            raise ModeNotFoundError(octave_channel)
        else:
            return query_result.doc_id

    def _lo_mode_id(self, mode_id: int, lo_freq: Number, gain: Optional[float], create: bool = False) -> int:
        query_result = self._query_lo_mode(mode_id, lo_freq, gain)
        if query_result is None:
            if create:
                return self._db.table("lo_modes").insert(asdict(_LOMode(mode_id, lo_freq, gain, 0)))
            else:
                raise ModeNotFoundError((mode_id, lo_freq))
        else:
            return query_result.doc_id

    def _if_mode_id(self, lo_mode_id: int, if_freq: float, create: bool = False) -> int:
        query_result = self.query_if_mode(lo_mode_id, if_freq)
        if query_result is None:
            if create:
                return self._db.table("if_modes").insert(asdict(_IFMode(lo_mode_id, if_freq, 0)))
            else:
                raise ModeNotFoundError((lo_mode_id, if_freq))
        else:
            return query_result.doc_id

    def update_lo_calibration(
        self,
        octave_channel: Tuple[str, int],
        lo_freq: float,
        gain: Optional[float],
        i0: float,
        q0: float,
        dc_gain: float,
        dc_phase: float,
        temperature: float,
        method: str = "",
    ) -> None:

        mode_id = self._mode_id(octave_channel, create=True)
        lo_mode_id = self._lo_mode_id(mode_id, lo_freq, gain, create=True)

        timestamp = time.time()

        lo_cal_id = self._db.table("lo_cal").insert(
            asdict(LOCalibrationDBSchema(i0, q0, lo_mode_id, dc_gain, dc_phase, temperature, timestamp, method))
        )

        self._db.table("lo_modes").update(asdict(_LOMode(mode_id, lo_freq, gain, lo_cal_id)), doc_ids=[lo_mode_id])

    def update_if_calibration(
        self,
        octave_channel: Tuple[str, int],
        lo_freq: float,
        output_gain: Optional[float],
        if_freq: float,
        gain: float,
        phase: float,
        temperature: float,
        method: str = "",
    ) -> None:

        mode_id = self._mode_id(octave_channel, create=True)
        lo_mode_id = self._lo_mode_id(mode_id, lo_freq, output_gain, create=True)
        if_mode_id = self._if_mode_id(lo_mode_id, if_freq, create=True)

        timestamp = time.time()
        if_cal_id = self._db.table("if_cal").insert(
            asdict(IFCalibrationDBSchema(if_mode_id, gain, phase, temperature, timestamp, method))
        )

        self._db.table("if_modes").update(asdict(_IFMode(lo_mode_id, if_freq, if_cal_id)), doc_ids=[if_mode_id])

    def get_lo_cal(
        self, octave_channel: Tuple[str, int], lo_freq: float, gain: Optional[float]
    ) -> Optional[LOCalibrationDBSchema]:
        try:
            mode_id = self._mode_id(octave_channel)
        except ModeNotFoundError:
            return None
        lo_mode = self._query_lo_mode(mode_id, lo_freq, gain)
        if lo_mode is None or lo_mode["latest"] == 0:
            return None

        lo_cal = cast(Optional[Document], self._db.table("lo_cal").get(doc_id=lo_mode["latest"]))
        if lo_cal is None:
            return None

        lo_cal_obj = LOCalibrationDBSchema(**lo_cal)

        return lo_cal_obj

    def get_if_cal(
        self, octave_channel: Tuple[str, int], lo_freq: Number, gain: Optional[float], if_freq: Number
    ) -> Optional[IFCalibrationDBSchema]:
        try:
            mode_id = self._mode_id(octave_channel, create=False)
            lo_mode_id = self._lo_mode_id(mode_id, lo_freq, gain, create=False)
            if_mode = self.query_if_mode(lo_mode_id, if_freq)
        except ModeNotFoundError:
            return None

        if if_mode is None or if_mode["latest"] == 0:
            return None

        if_cal = self._db.table("if_cal").get(doc_id=if_mode["latest"])
        if if_cal is None:
            return None
        assert isinstance(if_cal, dict)
        if_cal_obj = IFCalibrationDBSchema(**if_cal)

        return if_cal_obj

    def get_all_if_cal_for_lo(
        self, octave_channel: Tuple[str, int], lo_freq: float, gain: Optional[float]
    ) -> Dict[Number, IFCalibrationDBSchema]:
        try:
            mode_id = self._mode_id(octave_channel)
            lo_mode_id = self._lo_mode_id(mode_id, lo_freq, gain)
        except ModeNotFoundError:
            return {}

        if_modes = self._db.table("if_modes").search(where("lo_mode_id") == lo_mode_id)

        if if_modes is None:
            return {}

        if_dict: Dict[Number, IFCalibrationDBSchema] = {}
        for if_mode in if_modes:
            if_freq = if_mode["if_freq"]
            if_cal = self.get_if_cal(octave_channel, lo_freq, gain, if_freq)

            if if_cal:
                if_dict[if_freq] = if_cal

        return if_dict

    def update_calibration_result(
        self, result: MixerCalibrationResults, octave_channel: Tuple[str, int], method: str = ""
    ) -> None:

        for (lo_freq, output_gain), lo_cal in result.items():
            self.update_lo_calibration(
                octave_channel,
                lo_freq,
                output_gain,
                lo_cal.i0,
                lo_cal.q0,
                lo_cal.dc_gain,
                lo_cal.dc_phase,
                lo_cal.temperature,
                method,
            )

            for if_freq, if_cal in lo_cal.image.items():
                fine_cal = if_cal.fine
                self.update_if_calibration(
                    octave_channel,
                    lo_freq,
                    output_gain,
                    if_freq,
                    fine_cal.gain,
                    fine_cal.phase,
                    lo_cal.temperature,
                    method,
                )



from qm.utils.deprecation_utils import throw_warning
from qm.jobs.job_queue_old_api import QmQueue  # noqa

throw_warning(
    "'qm.QmQueue.QmQueue' is moved as of 1.1.0 and will be removed in 1.2.0. " "use 'qm.QmQueue' instead",
    category=DeprecationWarning,
    stacklevel=2,
)



import warnings
from enum import Enum
from typing import Tuple

from qm.jobs.base_job import QmBaseJob
from qm.utils import deprecation_message
from qm.grpc.general_messages import Matrix
from qm.results import StreamingResultFetcher
from qm.api.job_result_api import JobResultServiceApi
from qm._report import ExecutionError, ExecutionReport


class AcquiringStatus(Enum):
    AcquireStopped = 0
    NoDataToAcquire = 1
    HasDataToAcquire = 2


class RunningQmJob(QmBaseJob):
    @property
    def manager(self) -> None:
        """The QM object where this job lives

        Returns:

        """
        warnings.warn(
            deprecation_message(
                method="RunningQmJob.manager",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="QMJob no longer has 'manager' property",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return None

    @property
    def result_handles(self) -> StreamingResultFetcher:
        """:type: qm._results.JobResults

        Returns:
            The handles that this job generated
        """
        return StreamingResultFetcher(
            self._id,
            JobResultServiceApi(self._frontend.connection_details, self._id),
            self._store,
            self._capabilities,
        )

    def cancel(self) -> bool:
        return self.halt()

    def halt(self) -> bool:
        """Halts the job on the opx"""
        return self._job_manager.halt(self._id)

    def resume(self) -> bool:
        """Resumes a program that was halted using the [qm.qua._dsl.pause][] statement"""
        return self._job_manager.resume(self._id)

    def is_paused(self) -> bool:
        """Returns:
            Returns ``True`` if the job is in a paused state.

        see also:

            ``resume()``
        """
        return self._job_manager.is_paused(self._id)

    def _is_job_running(self) -> bool:
        """Returns:
        Returns ``True`` if the job is running
        """
        return self._job_manager.is_job_running(self._id)

    def _is_data_acquiring(self) -> AcquiringStatus:
        """Returns the data acquiring status.
        The possible statuses are: AcquireStopped, NoDataToAcquire,  HasDataToAcquire

        Returns:
            An AcquiringStatus enum object
        """
        status = self._job_manager.is_data_acquiring(self._id)
        return AcquiringStatus(status.value)

    def execution_report(self) -> ExecutionReport:
        """Get runtime errors report for this job. See [Runtime errors](../../Guides/error/#runtime-errors).

        Returns:
            An object holding the errors that this job generated.
        """
        errors = [
            ExecutionError.create_from_grpc_message(item)
            for item in JobResultServiceApi(self._frontend.connection_details, self.id).get_job_errors()
        ]
        return ExecutionReport(self._id, errors)

    def set_element_correction(
        self, element: str, correction: Tuple[float, float, float, float]
    ) -> Tuple[float, float, float, float]:
        r"""Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with a element.

        Changes will only be done to the current job!

        Values will be rounded to an accuracy of $2^{-16}$.
        Valid values for the correction values are between $-2$ and $(2 - 2^{-16})$.

        Warning - the correction matrix can increase the output voltage which might result in an
        overflow.

        Args:
            element (str): the name of the element to update the
                correction for
            correction (tuple):

                tuple is of the form (v00, v01, v10, v11) where
                the matrix is
                $\begin{pmatrix} v_{00} & v_{01} \\ v_{10} & v_{11}\end{pmatrix}$

        Returns:
            The correction matrix, after rounding to the OPX resolution.
        """
        correction_matrix = Matrix(v00=correction[0], v01=correction[1], v10=correction[2], v11=correction[3])

        return self._job_manager.set_element_correction(self._id, element, correction_matrix)

    def get_element_correction(self, element: str) -> Tuple[float, float, float, float]:
        """Gets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with a element.

        Args:
            element (str): the name of the element to update the
                correction for

        Returns:
            The current correction matrix
        """
        return self._job_manager.get_element_correction(self._id, element)



from qm.utils.deprecation_utils import throw_warning
from qm.serialization.qua_node_visitor import QuaNodeVisitor  # noqa

throw_warning(
    "'qm.QuaNodeVisitor.QuaNodeVisitor' is moved as of 1.1.0 and will be removed in 1.2.0. "
    "use 'qm.serialization.qua_serializing_visitor.QuaNodeVisitor' instead",
    category=DeprecationWarning,
    stacklevel=2,
)



from qm.utils.deprecation_utils import throw_warning
from qm.api.models.capabilities import ServerCapabilities  # noqa

throw_warning(
    "'qm.capabilities.ServerCapabilities' is moved as of 1.1.0 and will be removed in 1.2.0. "
    "use 'qm.api.models.capabilities.ServerCapabilities' instead",
    category=DeprecationWarning,
)



import logging
from typing import List, Type, Tuple, Union, Literal, Mapping, Iterable, Optional, TypedDict, cast, overload

from qm.utils import LOG_LEVEL_MAP
from qm.octave import QmOctaveConfig
from qm.persistence import BaseStore
from qm.api.v2.job_api import JobStatus
from qm.grpc.qua_config import QuaConfig
from qm.utils.async_utils import run_async
from qm.api.v2.base_api_v2 import BaseApiV2
from qm.program import Program, load_config
from qm.grpc.compiler import CompilerMessage
from qm.elements_db import init_octave_elements
from qm.grpc.frontend import SimulatedResponsePart
from qm.octave.octave_manager import OctaveManager
from qm.simulate.interface import SimulationConfig
from qm.type_hinting.config_types import DictQuaConfig
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.program.ConfigBuilder import convert_msg_to_config
from qm.api.simulation_api import create_simulation_request
from qm.api.v2.job_api.simulated_job_api import SimulatedJobApi
from qm.elements.up_converted_input import UpconvertedInputNewApi
from qm.api.v2.job_api.job_api import JobApi, JobData, transfer_statuses_to_enum
from qm.exceptions import QopResponseError, CompilationException, FailedToExecuteJobException
from qm.octave.qm_octave import QmOctaveForNewApi, create_mixer_correction, create_dc_offset_octave_update
from qm.api.models.compiler import CompilerOptionArguments, standardize_compiler_params, get_request_compiler_options
from qm.octave.octave_mixer_calibration import (
    AutoCalibrationParams,
    MixerCalibrationResults,
    NewApiOctaveMixerCalibration,
)
from qm.grpc.v2 import (
    QmServiceStub,
    JobsQueryParams,
    SimulationError,
    RemoveJobsRequest,
    UpdateConfigRequest,
    QmServiceCloseRequest,
    QmServiceCompileRequest,
    QmServiceGetJobsRequest,
    QmServiceSimulateRequest,
    QmServiceGetConfigRequest,
    QmServiceAddToQueueRequest,
    CompileResponseCompilationError,
    QmServiceAddCompiledToQueueRequest,
    AddToQueueResponseAddToQueueResponseError,
)

logger = logging.getLogger(__name__)


def _log_messages(messages: List[CompilerMessage]) -> None:
    for message in messages:
        logger.log(LOG_LEVEL_MAP[message.level], message.message)


class QmApi(BaseApiV2[QmServiceStub]):
    SIMULATED_JOB_CLASS = SimulatedJobApi

    def __init__(
        self,
        connection_details: ConnectionDetails,
        qm_id: str,
        store: BaseStore,
        capabilities: ServerCapabilities,
        octave_config: Optional[QmOctaveConfig],
        octave_manager: OctaveManager,
        pb_config: Optional[QuaConfig] = None,
    ) -> None:
        # todo - remove _pb_config when octave config is in the GW
        super().__init__(connection_details)
        self._caps = capabilities
        self._id = qm_id
        self._store = store
        pb_config = pb_config or self._get_pb_config()
        self._elements = init_octave_elements(pb_config, octave_config)
        self._octave_manager = octave_manager
        self._octave = QmOctaveForNewApi(self, octave_manager)

    @property
    def octave(self) -> QmOctaveForNewApi:
        return self._octave

    @property
    def id(self) -> str:
        """Hopefully temporary, till we move to the new API."""
        return self._id

    @property
    def _stub_class(self) -> Type[QmServiceStub]:
        return QmServiceStub

    def update_config(self, config: DictQuaConfig) -> None:
        config_pb = load_config(config)
        request = UpdateConfigRequest(quantum_machine_id=self._id, config=config_pb)
        run_async(self._stub.update_config(request, timeout=self._timeout))

    def get_jobs(
        self,
        job_ids: Iterable[str] = tuple(),
        user_ids: Iterable[str] = tuple(),
        description: str = "",
        status: Union[JobStatus, Iterable[JobStatus]] = tuple(),
    ) -> List[JobData]:
        query_params = JobsQueryParams(
            quantum_machine_ids=[self._id],
            job_ids=list(job_ids),
            user_ids=list(user_ids),
            description=description,
            status=transfer_statuses_to_enum(status),
        )
        request = QmServiceGetJobsRequest(query=query_params)
        response = self._run(self._stub.get_jobs(request, timeout=self._timeout))
        return [JobData.from_grpc(j) for j in response.jobs]

    def _get_pb_config(self) -> QuaConfig:
        request = QmServiceGetConfigRequest(quantum_machine_id=self._id)
        response = self._run(self._stub.get_config(request, timeout=self._timeout))
        return response.config

    def get_config(self) -> DictQuaConfig:
        return convert_msg_to_config(self._get_pb_config())

    def compile(self, program: Program, compiler_options: Optional[CompilerOptionArguments] = None) -> str:
        if compiler_options is None:
            compiler_options = CompilerOptionArguments()
        program.qua_program.compiler_options = get_request_compiler_options(compiler_options)
        request = QmServiceCompileRequest(quantum_machine_id=self._id, high_level_program=program.qua_program)
        try:
            response = self._run(self._stub.compile(request, timeout=self._timeout))
        except QopResponseError as e:
            error: CompileResponseCompilationError = e.error
            error_messages = []
            logger.error("Compilation failed")
            for msg in error.messages:
                logger.error(msg)
                error_messages.append(msg.message)
            formatted_errors = "\n".join(error_messages)
            raise CompilationException("Compilation failed, see the following errors:\n" + formatted_errors)

        _log_messages(response.messages)
        return response.program_id

    def _add_compiled(self, program_id: str) -> JobApi:
        request = QmServiceAddCompiledToQueueRequest(quantum_machine_id=self._id, program_id=program_id)
        response = self._run(self._stub.add_compiled_to_queue(request, timeout=self._timeout))
        return self._get_job(response.job_id, store=self._store)

    def _add_program(self, program: Program) -> JobApi:
        request = QmServiceAddToQueueRequest(quantum_machine_id=self._id, high_level_program=program.qua_program)
        try:
            response = self._run(self._stub.add_to_queue(request, timeout=self._timeout))
        except QopResponseError as e:
            error: AddToQueueResponseAddToQueueResponseError = e.error
            error_messages = []
            logger.error(f"Execution failed - {error.details}")
            for msg in error.messages:
                logger.error(msg)
                error_messages.append(msg.message)
            formatted_errors = "\n".join(error_messages)

            _log_messages(e.error.messages)
            raise FailedToExecuteJobException(
                f"Failed to execute program. See the following errors:\n{formatted_errors}"
            )
        _log_messages(response.messages)
        return self._get_job(response.job_id, store=self._store)

    @overload
    def add_to_queue(self, program: str) -> JobApi:
        pass

    @overload
    def add_to_queue(self, program: Program, *, compiler_options: Optional[CompilerOptionArguments] = None) -> JobApi:
        pass

    def add_to_queue(
        self, program: Union[Program, str], *, compiler_options: Optional[CompilerOptionArguments] = None
    ) -> JobApi:
        logger.info("Adding program to queue.")
        if isinstance(program, str):
            if compiler_options:
                logger.warning("Cannot add compiler options to a compiled program. Ignoring.")
            return self._add_compiled(program)
        else:
            if compiler_options is None:
                compiler_options = CompilerOptionArguments()
            program.qua_program.compiler_options = get_request_compiler_options(compiler_options)
            return self._add_program(program)

    def get_job(self, job_id: str) -> JobApi:
        return self._get_job(job_id, self._store)

    def _get_job(self, job_id: str, store: BaseStore) -> JobApi:
        return JobApi(self.connection_details, job_id, store, self._caps)

    def _get_simulated_job(self, job_id: str, simulated: SimulatedResponsePart) -> SimulatedJobApi:
        return self.SIMULATED_JOB_CLASS(
            self.connection_details, job_id, store=self._store, simulated_response=simulated, capabilities=self._caps
        )

    def simulate(
        self,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: Optional[CompilerOptionArguments] = None,
        *,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> SimulatedJobApi:
        standardized_compiler_options = standardize_compiler_params(compiler_options, strict, flags)
        standard_request = create_simulation_request(
            self._get_pb_config(), program, simulate, standardized_compiler_options
        )
        request = QmServiceSimulateRequest(
            quantum_machine_id=self._id,
            high_level_program=standard_request.high_level_program,
            simulate=standard_request.simulate,
            controller_connections=standard_request.controller_connections,
        )
        logger.info("Simulating program.")
        try:
            response = self._run(self._stub.simulate(request, timeout=self._timeout))
        except QopResponseError as e:
            raise handle_simulation_error(e)

        _log_messages(response.messages)

        return self._get_simulated_job(response.job_id, response.simulated)

    def execute(self, program: Program, *, compiler_options: Optional[CompilerOptionArguments] = None) -> JobApi:
        return self.add_to_queue(program, compiler_options=compiler_options)

    def clear_queue(
        self,
        job_ids: Iterable[str] = tuple(),
        user_ids: Iterable[str] = tuple(),
        description: str = "",
        status: Union[JobStatus, Iterable[JobStatus]] = tuple(),
    ) -> int:
        query_params = JobsQueryParams(
            quantum_machine_ids=[self._id],
            job_ids=list(job_ids),
            user_ids=list(user_ids),
            description=description,
            status=transfer_statuses_to_enum(status),
        )
        response = self._run(self._stub.remove_jobs(RemoveJobsRequest(query_params), timeout=self._timeout))
        return len(response.removed_job_ids)

    def close(self) -> None:
        self._run(self._stub.close(QmServiceCloseRequest(quantum_machine_id=self._id), timeout=self._timeout))

    def get_queue_count(self) -> int:
        jobs = self.get_jobs(status=["In queue"])
        return len(jobs)

    def calibrate_element(
        self,
        qe: str,
        lo_if_dict: Optional[Mapping[float, Tuple[float, ...]]] = None,
        save_to_db: bool = True,
        params: Optional[AutoCalibrationParams] = None,
    ) -> MixerCalibrationResults:

        inst = self._elements[qe]

        if params is None:
            params = AutoCalibrationParams()

        inst_input = inst.input
        assert isinstance(inst_input, UpconvertedInputNewApi)

        if lo_if_dict is None:
            lo_if_dict = {inst_input.lo_frequency: (inst.intermediate_frequency,)}
        client = self._octave_manager._get_client_from_port(inst_input.port)
        res = NewApiOctaveMixerCalibration(client=client, qm_api=self).calibrate(
            element=inst,
            lo_if_dict=lo_if_dict,
            params=params,
        )

        if save_to_db:
            calibration_db = self._octave_manager._octave_config._calibration_db
            if calibration_db is None:
                logger.warning("No calibration db found, can't save results")
            else:
                calibration_db.update_calibration_result(res, inst_input.port, "auto")

        key = (inst_input.lo_frequency, cast(float, inst_input.gain))
        if key in res:
            qe_cal = res[key]
            if inst.intermediate_frequency in qe_cal.image:
                update = create_dc_offset_octave_update(inst_input, i_offset=qe_cal.i0, q_offset=qe_cal.q0)
            else:
                update = {"version": 1}
            mixers = [
                create_mixer_correction(if_freq, inst_input.lo_frequency, if_cal.fine.correction)
                for if_freq, if_cal in qe_cal.image.items()
            ]
            update["mixers"] = {inst_input.mixer: mixers}
            self.update_config(update)
        return res


def handle_simulation_error(e: QopResponseError[SimulationError]) -> FailedToExecuteJobException:
    error = e.error
    logger.error("Job failed. Failed to execute program.")

    error_messages = []
    for msg in error.messages:
        lvl = LOG_LEVEL_MAP[msg.level]
        if lvl == logging.ERROR:
            error_messages.append(msg.message)
        logger.log(lvl, msg.message)

    formatted_errors = "\n".join(error_messages)

    return FailedToExecuteJobException(
        f"Job failed. Failed to execute program. see the following errors:\n{formatted_errors}"
    )


class NoRunningQmJob(Exception):
    pass


IoValue = TypedDict(
    "IoValue",
    {"io_number": Literal[1, 2], "int_value": int, "fixed_value": float, "boolean_value": bool},
    total=False,
)



from qm.utils.deprecation_utils import throw_warning
from qm.results import StreamingResultFetcher, SingleStreamingResultFetcher, MultipleStreamingResultFetcher  # noqa

# Backward compatible names
JobResults = StreamingResultFetcher
SingleNamedJobResult = SingleStreamingResultFetcher
MultipleNamedJobResult = MultipleStreamingResultFetcher

throw_warning(
    "'qm._results.JobResults' is moved as of 1.1.0 and will be removed in 1.2.0. "
    "use 'qm.results.JobResults' instead",
    category=DeprecationWarning,
    stacklevel=2,
)



from qm.jobs.pending_job import QmPendingJob  # noqa
from qm.utils.deprecation_utils import throw_warning

throw_warning(
    "'qm.QmPendingJob.QmPendingJob' is moved as of 1.1.0 and will be removed in 1.2.0. "
    "use 'qm.QmPendingJob' instead",
    category=DeprecationWarning,
    stacklevel=2,
)



from qm.jobs.qm_job import QmJob  # noqa
from qm.utils.deprecation_utils import throw_warning

throw_warning(
    "'qm.QmJob.QmJob' is moved as of 1.1.0 and will be removed in 1.2.0. " "use 'qm.QmJob' instead",
    category=DeprecationWarning,
    stacklevel=2,
)
